import git

import base64
import itertools
import pickle
import os
import collections


PREFIX = "B2LUIGI_PARAM_"


def product_dict(**kwargs):
    keys = kwargs.keys()
    vals = kwargs.values()
    for instance in itertools.product(*vals):
        yield dict(zip(keys, instance))


def fill_kwargs_with_lists(**kwargs):
    return_kwargs = {}
    for key, value in kwargs.items():
        if value is None:
            value = []
        if not isinstance(value, list):
            value = [value]
        return_kwargs[key] = value

    return return_kwargs


def encode_value(value):
    pickled_value = pickle.dumps(value)
    encoded_bytes = base64.b64encode(pickled_value)
    encoded_string = encoded_bytes.decode()
    return encoded_string


def decode_value(encoded_string):
    encoded_bytes = encoded_string.encode()
    pickled_value = base64.b64decode(encoded_bytes)
    value = pickle.loads(pickled_value)
    return value


def get_basf2_git_hash():
    basf2_release = os.getenv("BELLE2_RELEASE")

    assert basf2_release

    if basf2_release == "head":
        basf2_release_location = os.getenv("BELLE2_LOCAL_DIR")

        assert basf2_release_location
        return git.Repo(basf2_release_location).head.object.hexsha

    return basf2_release


def flatten_to_file_paths(inputs):
    if not inputs:
        return None

    return {key: value.path for key, value in inputs.items()}

def flatten_to_dict(inputs):
    if isinstance(inputs, dict):
        return inputs

    if isinstance(inputs, collections.Iterable):
        joined_dict = {}
        for i in inputs:
            joined_dict.update(**i)
        return joined_dict

    return {inputs: inputs}


def task_iterator(task):
    yield task
    for dep in task.deps():
        yield from task_iterator(dep)


from b2luigi.cli.arguments import get_cli_arguments
from b2luigi.cli.runner import run_as_batch_worker, run_local, run_test_mode, run_batched
from b2luigi.core import tasks, helper_tasks, utils

import basf2

import inspect
import os


############################################################################
# TODO: remove all this
def show_all_outputs(task_list):
    pass


def get_tasks_from_path_creator_function(path_creator_function, kwargs):
    parameters = inspect.signature(path_creator_function).parameters

    for key, param in parameters.items():
        setattr(helper_tasks.Basf2PathCreatorTask, key, tasks.NonParseableParameter())

    refined_kwargs = {}
    for key, param in parameters.items():
        if param.default == inspect.Parameter.empty:
            value = kwargs.pop(key)
        else:
            value = kwargs.pop(key, param.default)

        refined_kwargs[key] = value

    product_dict = utils.product_dict(**refined_kwargs)

    return [helper_tasks.Basf2PathCreatorTask(path_creator_function=path_creator_function, **path_kwargs) for path_kwargs in product_dict]


global_basf2_path = None


def create_path():
    return global_basf2_path


def get_tasks_from_basf2_path(basf2_path, kwargs):
    max_event = kwargs.pop("max_event", 0)

    helper_tasks.Basf2PathTask.basf2_path = basf2_path
    return [helper_tasks.Basf2PathTask(max_event=max_event)]
############################################################################


__has_run_already = False


def process(task_like_elements, **kwargs):
    # Assert, that process is only run once
    global __has_run_already
    if __has_run_already:
        raise RuntimeError("You are not allowed to call process twice in your code!")
    __has_run_already = True

    # Create Task List
    if isinstance(task_like_elements, basf2.Path):
        task_list = get_tasks_from_basf2_path(task_like_elements, kwargs)
    elif callable(task_like_elements):
        task_list = get_tasks_from_path_creator_function(task_like_elements, kwargs)
    elif not isinstance(task_like_elements, list):
        task_list = [task_like_elements]
    else:
        task_list = task_like_elements

    # Run now if requested
    if os.environ.get("B2LUIGI_EXECUTION", False):
        return tasks.run_task_from_env()

    # Check the CLI arguments and run as requested
    cli_args = get_cli_arguments()

    if cli_args.show_output:
        show_all_outputs(task_list)
    elif cli_args.test:
        run_test_mode(task_list, cli_args, kwargs)
    elif cli_args.batch:
        run_batched(task_list, cli_args, kwargs)
    elif cli_args.batch_runner:
        run_as_batch_worker(task_list, cli_args, kwargs)
    else:
        run_local(task_list, cli_args, kwargs)






from b2luigi import ROOTLocalTarget
from b2luigi.core import tasks

import luigi
import basf2
import ROOT

import inspect
import subprocess


class Basf2Task(tasks.DispatchableTask):
    num_processes = luigi.IntParameter(significant=False, default=0)
    max_event = luigi.IntParameter(significant=False, default=0)

    def create_path(self):
        raise NotImplementedError()

    def process(self):
        if self.num_processes:
            basf2.set_nprocesses(self.num_processes)

        if self.max_event:
            ROOT.Belle2.Environment.Instance().setNumberEventsOverride(self.max_event)

        path = self.create_path()

        basf2.print_path(path)
        basf2.process(path)

        print(basf2.statistics)


class SimplifiedOutputBasf2Task(Basf2Task):
    def output(self):
        path = self.create_path()
        outputs = []

        for module in path.modules():
            if module.type() == "RootOutput":
                for param in module.available_params():
                    if param.name == "outputFileName":
                        outputs.append(ROOTLocalTarget(param.values))

        return outputs


class Basf2PathCreatorTask(SimplifiedOutputBasf2Task):
    path_creator_function = tasks.FunctionParameter()

    def create_path(self):
        path_creator_function = self.path_creator_function
        parameters = inspect.signature(path_creator_function).parameters

        path_function_parameters = {}
        for key, value in self.get_filled_params().items():
            if key in parameters:
                path_function_parameters[key] = value

        return path_creator_function(**path_function_parameters)


class Basf2PathTask(SimplifiedOutputBasf2Task):
    basf2_path = None
    
    def create_path(self):
        return self.basf2_path


class HaddTask(tasks.Task):
    def output(self):
        output_targets = {}

        for key, _ in self.get_transposed_input_file_names().items():
            if hasattr(self, "keys") and key not in self.keys:
                continue

            yield self.add_to_output(key)

    def run(self):
        self.create_output_dirs()

        for key, file_list in self.get_transposed_input_file_names().items():
            if hasattr(self, "keys") and key not in self.keys:
                continue

            args = ["hadd", "-f", self.get_output_file_name(key)] + file_list
            subprocess.check_call(args)


class Basf2FileMergeTask(HaddTask):
    def run(self):
        self.create_output_dirs()

        for key, file_list in self.get_transposed_input_file_names().items():
            if hasattr(self, "keys") and key not in self.keys:
                continue
            args = ["b2file-merge", "-f", self.get_output_file_name(key)] + file_list
            subprocess.check_call(args)


import json
import os


# The global object hosting the current settings
_current_global_settings = {}


def get_setting(key, default=None):
    try:
        return _current_global_settings[key]
    except KeyError:
        for settings_file in _setting_file_iterator():
            try:
                with open(settings_file , "r") as f:
                    return json.load(f)[key]
            except KeyError:
                pass

    if default is None:
        raise ValueError(f"No settings found for {key}!")
    return default


def set_setting(key, value):
    _current_global_settings[key] = value


def _setting_file_iterator():
    path = os.getcwd()

    while True:
        json_file = os.path.join(path, "settings.json")
        if os.path.exists(json_file):
            yield json_file

        path = os.path.split(path)[0]
        if path == "/":
            break



import time

import b2luigi
import basf2
import luigi

class GenerationTask(b2luigi.SimplifiedOutputBasf2Task):
    number = luigi.IntParameter()

    def create_path(self):
        path = basf2.create_path()

        path.add_module("EventInfoSetter", evtNumList=[self.number])
        path.add_module("RootOutput", outputFileName=f"generated_{self.number}.root")

        return path

    def run_local(self):
        super().run_local()
        time.sleep(10)


class Killer(basf2.Module):
    def event(self):
        raise ValueError


class ReconstructionTask(b2luigi.SimplifiedOutputBasf2Task):
    number = luigi.IntParameter()

    def requires(self):
        return self.clone(GenerationTask)

    def create_path(self):
        path = basf2.create_path()

        path.add_module("RootInput", inputFileNames=self.get_input_file_names())

        if self.number == 3:
            path.add_module(Killer())

        path.add_module("RootOutput", outputFileName=f"reconstruction_{self.number}.root")

        return path

    def run_local(self):
        super().run_local()
        time.sleep(10)


class MainTask(luigi.WrapperTask):
    def requires(self):
        return [ReconstructionTask(number=1), ReconstructionTask(number=3)]


if __name__ == "__main__":
    b2luigi.process(MainTask())



import json

import luigi
import luigi.server
import luigi.configuration

from b2luigi.cli.batch import SendJobWorkerSchedulerFactory
from b2luigi.core.settings import set_setting
from b2luigi.core.utils import task_iterator


def run_as_batch_worker(task_list, cli_args, kwargs):
    for root_task in task_list:
        for task in task_iterator(root_task):
            if task.task_id != cli_args.task_id:
                continue

            # TODO: We do not process the information if (a) we have a new dependency and (b) why the task has failed.
            try:
                task.run()
                task.on_success()
            except BaseException as ex:
                task.on_failure(ex)
                raise ex


def run_batched(task_list, cli_args, kwargs):
    kwargs["worker_scheduler_factory"] = SendJobWorkerSchedulerFactory()
    run_local(task_list, cli_args, kwargs)


def run_local(task_list, cli_args, kwargs):
    if cli_args.scheduler_host or cli_args.scheduler_port:
        core_settings = luigi.interface.core()
        host = cli_args.scheduler_host or core_settings.scheduler_host
        port = int(cli_args.scheduler_port) or core_settings.scheduler_port
        luigi.build(task_list, log_level="INFO", scheduler_host=host, scheduler_port=port, **kwargs)
    else:
        luigi.build(task_list, log_level="INFO", local_scheduler=True, **kwargs)


def run_test_mode(task_list, cli_args, kwargs):
    set_setting("dispatch", False)
    luigi.build(task_list, log_level="DEBUG", local_scheduler=True, **kwargs)



import json
import os
import re
import sys
import time

import luigi
import luigi.interface

import subprocess

from b2luigi.core.settings import get_setting

import enum


class BatchSystems(enum.Enum):
    lsf = "lsf"
    htcondor = "htcondor"
    local = "local"


class BatchProcess:
    def __init__(self, result_queue, worker_timeout, task, scheduler):
        self.use_multiprocessing = False
        self.exitcode = 0
        self.task = task
        self.timeout_time = time.time() + worker_timeout if worker_timeout else None

        self._result_queue = result_queue
        self._scheduler = scheduler

        self.task_cmd = [sys.executable, os.path.realpath(sys.argv[0]),
                         "--batch-runner",
                         "--task-id", task.task_id]

    def run(self):
        raise NotImplementedError

    def is_alive(self):
        raise NotImplementedError

    def terminate(self):
        raise NotImplementedError

    def put_to_result_queue(self):
        expl = ""
        missing = []
        new_deps = []
        self._result_queue.put((self.task.task_id, luigi.scheduler.DONE, expl, missing, new_deps))


class LSFProcess(BatchProcess):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._batch_job_id = None

    def is_alive(self):
        assert self._batch_job_id

        output = subprocess.check_output(["bjobs", "-json", "-o", "stat exit_code", self._batch_job_id])
        output = output.decode()
        output = json.loads(output)["RECORDS"][0]

        if "STAT" not in output:
            self.exitcode = -1
            return False

        job_status = output["STAT"]
        self.exitcode = output["EXIT_CODE"]

        if job_status == "DONE":
            self.put_to_result_queue()
            return False
        elif job_status == "EXIT":
            return False

        return True

    def run(self):
        prefix = ["bsub", "-env all"]

        try:
            prefix += ["-q", self.task.queue]
        except AttributeError:
            pass

        # Automatic requeing?

        try:
            stdout_log_file = self.task.log_files["stdout"]
            stderr_log_file = self.task.log_files["stderr"]
            prefix += ["-eo", stderr_log_file, "-oo", stdout_log_file]
        except AttributeError:
            pass

        output = subprocess.check_output(prefix + self.task_cmd)
        output = output.decode()

        # Output of the form Job <72065926> is submitted to default queue <s>.
        match = re.search(r"<[0-9]+>", output)
        if not match:
            raise RuntimeError("Batch submission failed with output " + output)

        self._batch_job_id = match.group(0)[1:-1]

    def terminate(self):
        if not self._batch_job_id:
            return

        subprocess.check_call(["bkill", self._batch_job_id], stdout=subprocess.DEVNULL)


class SendJobWorker(luigi.worker.Worker):
    def _create_task_process(self, task):
        batch_system = get_setting("batch_system", BatchSystems.lsf)

        # TODO: also use the task to choose the batch system!

        if batch_system == BatchSystems.lsf:
            return LSFProcess(task=task, scheduler=self._scheduler,
                              result_queue=self._task_result_queue, worker_timeout=self._config.timeout)
        elif batch_system == BatchSystems.htcondor:
            raise NotImplementedError
        else:
            return super()._create_task_process(task)


class SendJobWorkerSchedulerFactory(luigi.interface._WorkerSchedulerFactory):
    def create_worker(self, scheduler, worker_processes, assistant=False):
        return SendJobWorker(scheduler=scheduler, worker_processes=worker_processes, assistant=assistant)



"""Task scheduling and batch running for basf2 jobs made simple"""
from setuptools_scm import get_version
__version__ = get_version(root='..', relative_to=__file__)

from luigi import *

from luigi.util import requires, inherits

from b2luigi.core.tasks import Task, ExternalTask, WrapperTask, DispatchableTask, ROOTLocalTarget
from b2luigi.core.helper_tasks import Basf2Task, SimplifiedOutputBasf2Task, Basf2FileMergeTask, HaddTask
from b2luigi.cli.process import process
